/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2010 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.property;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.control.NameController;
import org.ow2.frascati.tinfi.control.content.SCAContentController;

/**
 * Mixin layer for implementing the {@link SCAContentController} interface for
 * scaPrimitive components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class SCAPropertyControllerMixin
implements SCAPropertyController {
    
    // -------------------------------------------------------------------------
    // Private constructor
    // -------------------------------------------------------------------------

    private SCAPropertyControllerMixin () {}

    // -------------------------------------------------------------------------
    // Implementation of the SCAPropertyController interface
    // -------------------------------------------------------------------------

    private Map<String,Object> values = new HashMap<String,Object>();
    private Map<String,Class<?>> types = new HashMap<String,Class<?>>();
    private Map<String,SCAPropertyController> promoters =
        new HashMap<String,SCAPropertyController>();
    
    /**
     * Set the type of the specified property. If the property type has already
     * been set, the old value is lost, and the new one is recorded.
     * 
     * The rationale for recording property types lies in the fact that complex
     * properties may have a type which differs from the one of the field where
     * their values will be injected. Typically, complex property values will be
     * injected in fields of type Object.
     * 
     * @param name   the property name
     * @param type   the property type
     * @since 1.1.1
     */
    public void setType( String name, Class<?> type ) {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            types.put(name,type);
        }
        else {
            promoter.setType(name,type);
        }
    }
    
    /**
     * Set the value of the specified property. If the property value has
     * already been set, the old value is lost, and the new one is recorded.
     * 
     * @param name   the property name
     * @param value  the property value
     * @throws IllegalArgumentException
     * 		if a type has been defined for the specified property with {@link
     * 		#setValue(String, Object)} and if the specified value is not an
     * 		instance of this type
     */
    public void setValue( String name, Object value )
    throws IllegalArgumentException {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {

        	// Check whether the value is of the correct type
        	if( types.containsKey(name) ) {
        		Class<?> type = types.get(name);
        		boolean isInstance = type.isInstance(value);
        		if( ! isInstance ) {
        			String msg =
        				"The type of the specified value "+value+
        				" is not assignable to the type "+type.getName()+
        				" previously defined with setType().";
        			throw new IllegalArgumentException(msg);
        		}
        	}
        	
        	// Record the new value
        	values.put(name,value);
            
            // Re-inject the value on current content instances
            _this_weaveableSCACC.setPropertyValue(name,value);
        }
        else {
            promoter.setValue(name,value);
        }
    }
    
    /**
     * Return the type of the specified property. Return <code>null</code> if
     * the property type has not been set.
     * 
     * @param name  the property name
     * @return      the property value
     * @since 1.1.1
     */
    public Class<?> getType( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        Class<?> type = 
        	promoter == null ?
            types.get(name) :
            promoter.getType(name);
        return type;
    }
    
    /**
     * Return the value of the specified property. Return <code>null</code> if
     * the property value has not been set.
     * 
     * @param name  the property name
     * @return      the property value
     */
    public Object getValue( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        Object value =
        	promoter == null ?
            values.get(name) :
            promoter.getValue(name);
        return value;
    }
    
    /**
     * Return <code>true</code> if the specified property has been set.
     * 
     * @param name  the property name
     * @return      <code>true</code> if the property has been set,
     *              <code>false</code> otherwise
     */
    public boolean containsPropertyName( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        boolean b =
	        promoter == null ?
	        values.containsKey(name) :
	        promoter.containsPropertyName(name);
        return b;
    }

    /**
     * Return the names of the properties whose values have been set by invoking
     * {@link #setValue(String, Object)}.
     * 
     * @since 1.1.1
     */
    public String[] getPropertyNames() {        
        Set<String> keys = values.keySet();
        String[] names = keys.toArray( new String[keys.size()] );
        return names;
    }
    
    /**
     * Return <code>true</code> if the specified property can be injected in the
     * content class.
     * 
     * @param name  the property name
     * @return      <code>true</code> if the property can be injected,
     *              <code>false</code> otherwise
     * @since 1.1.2
     */
    public boolean containsDeclaredPropertyName( String name ) {
        boolean b = _this_weaveableSCACC.containsPropertyName(name);
        return b;
    }
    
    /**
     * Return the names of the properties which can be injected in the content
     * class.
     * 
     * @since 1.1.1
     */
    public String[] getDeclaredPropertyNames() {
        String[] names = _this_weaveableSCACC.getPropertyNames();
        return names;
    }
    
    /**
     * Return the type of the specified property, provided that this property
     * can be injected in the content class.
     * 
     * @param name  the property name
     * @return      the property type
     * @since 1.1.2
     */
    public Class<?> getDeclaredPropertyType( String name ) {
    	Class<?> type = _this_weaveableSCACC.getPropertyType(name);
    	return type;
    }
    
    /**
     * Set the reference of the property controller which promotes the specified
     * property to the current property controller.
     * 
     * @param name      the promoter property name
     * @param promoter  the promoter component or
     *                  <code>null</code> to unregister the promoter
     * @throws IllegalPromoterException
     *      thrown when attempting to set a cycle between property promoters
     * @since 0.4.3
     */
    public void setPromoter( String name, SCAPropertyController promoter )
    throws IllegalPromoterException {
        
        promoters.put(name,promoter);

        if( promoter != null ) {
            SCAPropertyController peer = promoter.getPromoter(name);
            if( peer == this ) {
                String compname = _this_weaveableOptNC.getFcName();
                throw new IllegalPromoterException(compname);
            }
        }
    }
    
    /**
     * Return the reference of the property controller which promotes the
     * specified property. Return <code>null</code> if the property is managed
     * locally by the current property controller.
     * 
     * @param name  the promoter property name
     * @return      the promoter component or <code>null</code>
     * @since 0.4.3
     */
    public SCAPropertyController getPromoter( String name ) {
        return promoters.get(name);
    }

    
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    /**
     * The {@link SCAContentController} interface of the component to which this
     * controller object belongs.
     */
    public SCAContentController _this_weaveableSCACC;

    /**
     * The {@link NameController} interface of the component to which this
     * controller object belongs.
     */
    public NameController _this_weaveableOptNC;
}
